#include <stdio.h>
#include <math.h>
#include <gl.h>
#include <device.h>
#include "pwin.h"
#include "event.h"

int qiterate(float r, float i, float j, float k);

char *Progname;

void exit();
void draw_figure();

float R;
float I;
float J;
float K;
float Qr;
float Qi;
float Qj;
float Qk;
int Maxits = 100;

#define VARIABLE 100.0

float Left;
float Right;
float Bottom;
float Top;
int Wide;
int Tall;
int Maporigin;

int Init;

/* Parameter window stuff. */

PWIN *Pwin;

void get_float(int argc, char **argv, float *arg);
void get_int(int argc, char **argv, int *arg);
void set_pwin_items(PWIN *);
void save_pwin();

KEYWORD Keytable[] = {
	{ "r", get_float, &R },
	{ "i", get_float, &I },
	{ "j", get_float, &J },
	{ "k", get_float, &K },
	{ "qr", get_float, &Qr },
	{ "qi", get_float, &Qi },
	{ "qj", get_float, &Qj },
	{ "qk", get_float, &Qk },
	{ "Maxits", get_int, &Maxits },
	{ "Maporigin", get_int, &Maporigin },
	{ "left", get_float, &Left },
	{ "right", get_float, &Right },
	{ "bottom", get_float, &Bottom },
	{ "top", get_float, &Top },
	{ NULL, NULL, 0 }
};
#define NPWIN (sizeof(Keytable) / sizeof(KEYWORD) - 1)

int Win;

void redraw();
void hitme();
void drag_rect();
void reset();

main(int argc, char **argv)
{
	register char *curr_arg;
	register int c;
	int i;

	Progname = *argv;

	/* Parse arguments beginning with a hyphen. */

	while (--argc > 0 && **++argv == '-' && *(curr_arg = 1 + *argv)) {

		/* Scan current argument until a flag indicates that the rest
		of the argument isn't flags (curr_arg = NULL), or until
		the end of the argument is reached (if it is all flags). */

		while (curr_arg != NULL && (c = *curr_arg++)) {
			switch (c) {

			default:
				fprintf(stderr, "%s: unknown flag -%c\n", Progname, c);
				exit(-1);
			}
		}
	}

	/* Set up a window. */

	foreground();
	Win = winopen("");
	getsize(&Wide, &Tall);

	color(WHITE);
	clear();

	/* Parameter window. */

	R = I = VARIABLE;
	J = K = 0.0;

	Qr = -0.745;
	Qi = 0.113;
	Qj = 0.01;
	Qk = 0.01;

	Left = Bottom = -2.0;
	Right = Top = 2.0;

	Maporigin = 0;

	Pwin = create_pwin(NPWIN, 40, Keytable);

	/* If he specified a parameter file, read it in. */

	if (argc && **argv != '-') {
		scan_pfile(*argv, Pwin);
	}
	set_pwin_items(Pwin);

	/* Queue some events. */

	add_event(ANY, REDRAW, Win, redraw, 0);
	qdevice(REDRAW);
	add_event(ANY, ESCKEY, UP, exit, 0);
	qdevice(ESCKEY);
	add_event(ANY, WINQUIT, ANY, exit, 0);
	qdevice(WINQUIT);
	add_event(ANY, F1KEY, UP, save_pwin, 0);
	qdevice(F1KEY);
	add_event(ANY, F2KEY, UP, reset, 0);
	qdevice(F2KEY);

	/* Watch for mouse hits. */

	add_event(Win, LEFTMOUSE, DOWN, drag_rect, 0);
	qdevice(LEFTMOUSE);
	add_event(Win, MIDDLEMOUSE, DOWN, hitme, 0);
	qdevice(MIDDLEMOUSE);

	Init = TRUE;
	while (1) {
		if (Init) {
			Init = FALSE;
			draw_figure();
		} else {
			event();
		}
	}
}

/* Draw the figure. */

void
draw_figure()
{
	float *x, *y;
	float xinc, yinc;
	float r, i, j, k;
	int row, col;
	int its;

	ortho2(0, Wide, 0, Tall);
	color(WHITE);
	clear();

	r = R;
	i = I;
	j = J;
	k = K;

	if (r == VARIABLE) {
		x = &r;
	} else if (i == VARIABLE) {
		x = &i;
	} else {
		x = &j;
	}

	if (k == VARIABLE) {
		y = &k;
	} else if (j == VARIABLE) {
		y = &j;
	} else {
		y = &i;
	}

	xinc = (Right - Left) / Wide;
	yinc = (Top - Bottom) / Tall;
	for (col = 0; col < Wide; col++) {
		*x = Left + col * xinc;
		for (row = 0; row < Tall; row++) {
			*y = Bottom + row * yinc;
			its = qiterate(r, i, j, k);
			if (its == Maxits) {
				its = BLACK;
			}
			color(its + Maporigin);
			pnt2i(col, row);
		}
		if (qtest()) {
			event();
			if (Init) {
				break;
			}
		}
	}
}

/* Reset the items in the parameter window. */

void
set_pwin_items(PWIN *pwin)
{
	int i;
	KEYWORD *k;
	char buf[100];

	k = Keytable;
	for (i = 0; i < NPWIN; i++, k++) {
		if (k->keyfunc == (void (*)())get_float) {
			if (*((float *)k->arg) == VARIABLE) {
				sprintf(buf, "%s", k->keyword);
			} else {
				sprintf(buf, "%s %g", k->keyword, *((float *)k->arg));
			}
		} else if (k->keyfunc == (void (*)())get_int) {
			sprintf(buf, "%s %d", k->keyword, *((int *)k->arg));
		}
		add_pwin_item(pwin, i, buf);
	}
}

void
save_pwin()
{
	int i;
	char **pline;
	FILE *outfile;
	char buf[100];

	printf("File name? ");
	fgets(buf, 100, stdin);
	buf[strlen(buf) - 1] = '\0';
	if ((outfile = fopen(buf, "w")) == NULL) {
		fprintf(stderr, "can't write file %s\n", buf);
		return;
	}

	pline = Pwin->text;
	for (i = Pwin->n_items - 1; i >= 0; --i) {
		fputs(*pline++, outfile);
		putc('\n', outfile);
	}

	fclose(outfile);
}

/* Called from the parameter window when the user changes a parameter. */

void
get_float(int argc, char **argv, float *fp)
{
	if (argc == 1) {
		*fp = VARIABLE;
	} else {
		*fp = atof(argv[1]);
	}
	Init = TRUE;
}

void
get_int(int argc, char **argv, int *ip)
{
	*ip = atof(argv[1]);
	Init = TRUE;
}

void
reset()
{
	Left = Bottom = -2.0;
	Right = Top = 2.0;
	set_pwin_items(Pwin);
	Init = TRUE;
}

/* Quaternion addition and multiplication. */

typedef struct {
	float r;
	float i;
	float j;
	float k;
} QUAT;

void
qadd(QUAT *a, QUAT *b, QUAT *dst)
{
	dst->r = a->r + b->r;
	dst->i = a->i + b->i;
	dst->j = a->j + b->j;
	dst->k = a->k + b->k;
}

void
qmult(QUAT *a, QUAT *b, QUAT *dst)
{
	float r, i, j, k;

	r = a->r * b->r - a->i * b->i - a->j * b->j - a->k * b->k;
	i = a->r * b->i + a->i * b->r + a->j * b->k - a->k * b->j;
	j = a->r * b->j - a->i * b->k + a->j * b->r + a->k * b->i;
	k = a->r * b->k + a->i * b->j - a->j * b->i + a->k * b->r;

	dst->r = r;
	dst->i = i;
	dst->j = j;
	dst->k = k;
}

float
qpseudo_dist(QUAT *q)
{
	float f;

	f = q->r * q->r;
	f += q->i * q->i;
	f += q->j * q->j;
	f += q->k * q->k;

	return (f);
}

int
qclose(QUAT *q1, QUAT *q2)
{
	float f, x;

	x = q1->r - q2->r;
	f = x * x;
	x = q1->i - q2->i;
	f += x * x;
	x = q1->j - q2->j;
	f += x * x;
	x = q1->k - q2->k;
	f += x * x;

	return (f < 0.00001);
}

qprint(QUAT *q)
{
	printf("(%f, %f, %f, %f)\n", q->r, q->i, q->j, q->k);
}

#ifdef notdef
int
qiterate(float r, float i, float j, float k)
{
	QUAT q, c;
	int its;

	c.r = r;
	c.i = i;
	c.j = j;
	c.k = k;

	q.r = q.i = q.j = q.k = 0.0;
	for (its = 0; its < Maxits; its++) {
		qmult(&q, &q, &q);
		qadd(&q, &c, &q);
		if (qpseudo_dist(&q) > 4.0) {
			break;
		}
	}

	return (its);
}
#endif

int
qiterate(float r, float i, float j, float k)
{
	QUAT q, q0, c;
	int its;

	c.r = Qr;
	c.i = Qi;
	c.j = Qj;
	c.k = Qk;

	q.r = r;
	q.i = i;
	q.j = j;
	q.k = k;

	for (its = 0; its < Maxits; its++) {
		qmult(&q, &q, &q);
		qadd(&q, &c, &q);
		if (qpseudo_dist(&q) > 4.0) {
			return (Maxits);
		}
	}

	/* Try to determine the period of the fixed point. */

	q0 = q;
	for (its = 0; its < Maxits; its++) {
		qmult(&q, &q, &q);
		qadd(&q, &c, &q);
		if (qclose(&q, &q0)) {
qprint(&q);
			break;
		}
	}

	if (fabs(q.r - 0.0339) < 0.0001) {
		return (1);
	}
	if (fabs(q.r - 0.0660) < 0.0001) {
		return (2);
	}
	if (fabs(q.r - 0.2790) < 0.0001) {
		return (3);
	}
	if (fabs(q.r + 0.4222) < 0.0001) {
		return (4);
	}
	return (its + 1);
}

/* Middlemouse reports the coordinate, leftmouse drags a rect. */

void
hitme()
{
	int xorg, yorg;
	float xinc, yinc;
	float x, y;

	getorigin(&xorg, &yorg);

	xinc = (Right - Left) / Wide;
	yinc = (Top - Bottom) / Tall;

	x = Left + (getvaluator(MOUSEX) - xorg) * xinc;
	y = Bottom + (getvaluator(MOUSEY) - yorg) * yinc;
	printf("%f, %f\n", x, y);
}

void
redraw()
{
	reshapeviewport();
	getsize(&Wide, &Tall);
	Init = TRUE;
}

void
drag_rect()
{
	long mx, my, orx, ory, omx, omy, cx, cy, bx, by, obx, oby;
	float corner1x, corner1y, corner2x, corner2y;
	short val;

	drawmode(OVERDRAW);
	mapcolor(0, 0, 0, 0);
	mapcolor(1, 255, 255, 255);

	qreset();

	getorigin(&orx, &ory);
	omx = mx = getvaluator(MOUSEX) - orx;
	omy = my = getvaluator(MOUSEY) - ory;

	obx = bx = cx = mx;
	oby = by = cy = my;

	corner1y = ((double) my / (double) Tall) * (Top - Bottom) + Bottom;
	corner1x = ((double) mx / (double) Wide) * (Right - Left) + Left;

	color(0);
	clear();

	while (!qtest()) {
		mx = getvaluator(MOUSEX) - orx;
		my = getvaluator(MOUSEY) - ory;

		if (mx == omx && my == omy) {
			continue;
		}

#ifdef notdef
		/* preserve aspect ratio */

		if (abs(mx - cx) * Tall / Wide > abs(my - cy)) {
			bx = mx;
			by = cy + abs(mx - cx) * sgn(my - cy) * Tall / Wide;
		} else {
			by = my;
			bx = cx + abs(my - cy) * sgn(mx - cx) * Wide / Tall;
		}
#endif
		bx = mx;
		by = my;

		color(0);
		recti(cx, cy, obx, oby);

		color(1);
		recti(cx, cy, bx, by);

		obx = bx;
		oby = by;
		omx = mx;
		omy = my;

	}

	if (qtest() != LEFTMOUSE) {
		return;
	} else {
		qread(&val);
	}

	if (by != cy && bx != cx) {
		corner2y = (double) by / (double) Tall *(Top - Bottom) + Bottom;
		corner2x = (double) bx / (double) Wide *(Right - Left) + Left;

		if (corner2y < corner1y) {
			Bottom = corner2y;
			Top = corner1y;
		} else {
			Bottom = corner1y;
			Top = corner2y;
		}

		if (corner2x < corner1x) {
			Left = corner2x;
			Right = corner1x;
		} else {
			Left = corner1x;
			Right = corner2x;
		}
	}
	set_pwin_items(Pwin);
	Init = TRUE;

	color(0);
	clear();

	drawmode(NORMALDRAW);
}

sgn(i)
{
	return (i >= 0 ? (i == 0 ? 0 : 1) : -1);
}
