#include <stdio.h>
#include <math.h>

#define MAXITS 100

/* Quaternion addition and multiplication. */

typedef struct {
	float r;
	float i;
	float j;
	float k;
} QUAT;

void
qadd(QUAT *a, QUAT *b, QUAT *dst)
{
	dst->r = a->r + b->r;
	dst->i = a->i + b->i;
	dst->j = a->j + b->j;
	dst->k = a->k + b->k;
}

void
qmult(QUAT *a, QUAT *b, QUAT *dst)
{
	float r, i, j, k;

	r = a->r * b->r - a->i * b->i - a->j * b->j - a->k * b->k;
	i = a->r * b->i + a->i * b->r + a->j * b->k - a->k * b->j;
	j = a->r * b->j - a->i * b->k + a->j * b->r + a->k * b->i;
	k = a->r * b->k + a->i * b->j - a->j * b->i + a->k * b->r;

	dst->r = r;
	dst->i = i;
	dst->j = j;
	dst->k = k;
}

float
qpseudo_dist(QUAT *q)
{
	float f;

	f = q->r * q->r;
	f += q->i * q->i;
	f += q->j * q->j;
	f += q->k * q->k;

	return (f);
}

qprint(QUAT *q)
{
	printf("(%f, %f, %f, %f)\n", q->r, q->i, q->j, q->k);
}

int
iterate(float r, float i)
{
	QUAT q, c;
	int its;

	c.r = -0.745;
	c.i = 0.113;
	c.j = 0.01;
	c.k = 0.01;

	q.r = r;
	q.i = i;
	q.j = q.k = 0.025;
	for (its = 0; its < MAXITS; its++) {
		qmult(&q, &q, &q);
		qadd(&q, &c, &q);
		if (qpseudo_dist(&q) > 4.0) {
			break;
		}
	}

	return (its);
}

main()
{
	float r, i;
	int its;

	for (r = -2.0; r < 2.0; r += 0.008) {
		fprintf(stderr, "r = %f\n", r);
		for (i = -2.0; i < 2.0; i += 0.008) {
			its = iterate(r, i);
			if (its == MAXITS) {
				printf("%f %f\n", r, i);
			}
		}
	}
}
